from .parser import _Parser
from typing import Union, Generator


class DictFilter:
    """
    Constructs a DictFilter object, taking the SQL which will be applied to filter data
    :param sql: SQL Select statement which is used to filter data
    :raises InvalidTokenError: Raised when tokenising and an invalid value is read
    :raises UnexpectedTokenError: Raised when an unexpected token is encountered parsing the SQL
    """

    def __init__(self, sql: str):
        self._parser = _Parser(sql)

    """
    Applies the SQL provided when instantiated to a list, tuple of records or generator, returning those that match the criteria
    :param kwargs: Single named argument providing the collection to be filtered. The name of the argument must match the FROM reference in the SQL
    :returns: Collection of the same type provided, containing only the records in the source data that match the SQL criteria
    :raises: ValueError if parameters are invalid
    :raises: UnrecognisedReferenceError if a reference is made to a field not in the data
    """

    def filter(self, **kwargs) -> Union[list, tuple]:
        self._validate(**kwargs)
        coll_name = next(iter(kwargs.keys()))
        return (
            tuple(self._filter(kwargs[coll_name]))
            if isinstance(kwargs[coll_name], tuple)
            else self._filter(kwargs[coll_name])
        )

    """
    Applies the SQL provided when instantiated to a list, tuple of records or generator, yielding each machine record in turn
    :param kwargs: Single named argument providing the collection to be filtered. The name of the argument must match the FROM reference in the SQL
    :yields: Each record matching the SQL criteria
    :raises: ValueError if parameters are invalid
    :raises: UnrecognisedReferenceError if a reference is made to a field not in the data
    """

    def filtergen(self, **kwargs):
        self._validate(**kwargs)
        coll_name = next(iter(kwargs.keys()))
        for record in kwargs[coll_name]:
            if self._parser.satisfied(record):
                yield (self._parser.filter_fields(record))

    def _validate(self, **kwargs):
        if len(kwargs) != 1:
            raise ValueError(
                "Method takes one named parameter, denoting the data source"
            )
        coll_name = next(iter(kwargs.keys()))
        if coll_name != self._parser.from_ref():
            raise ValueError("Collection name does not match FROM reference in SQL")
        if not (
            isinstance(kwargs[coll_name], list) or isinstance(kwargs[coll_name], tuple) or isinstance(kwargs[coll_name], Generator)
        ):
            raise ValueError("Collection to be filtered must be a list, tuple or a generator")

    def _filter(self, source):
        return [
            self._parser.filter_fields(record)
            for record in source
            if self._parser.satisfied(record)
        ]



class InvalidTokenError(Exception):
    """
    Raised when tokenising and an invalid value is read
    """

    def __init__(self, token: str):
        super().__init__(f"Invalid token: {token}")


class UnexpectedTokenError(Exception):
    """
    Raised when parsing and an unexpected token is found
    """

    def __init__(self, found, expected=None):
        super().__init__(
            f"Unexpected token, expected {expected or 'end'}, found {found.ttype if found else 'no token'} with value {found.value if found else 'N/A'}."
        )


class UnrecognisedReferenceError(Exception):
    """
    Raised when executing SQL which includes a reference to a field which is not in the data
    """

    def __init__(self, reference):
        super().__init__(f"Unrecognised reference {reference}, not present in data.")



from collections import namedtuple
from enum import Enum
import re
from string import ascii_letters, digits
import typing

from pydictsql.exceptions import InvalidTokenError, UnexpectedTokenError

REFERENCE_PAT = re.compile(r"{[^}]*}")
NUMBER_PAT = re.compile(r"-?[0-9]+(\.[-0-9]+)?")
STRING_PAT = re.compile(r"(['\"])[^'\"]*\1")
SYMBOLS = "=<>(),*"

_Token = namedtuple("Token", ["ttype", "value"])


class _TokenType(Enum):
    REFERENCE = -1
    NUMBER = -2
    STRING = -3

    EQUALS = -4
    LT = -5
    LTE = -6
    GT = -7
    GTE = -8
    NE = -9
    LPAREN = -7
    RPAREN = -8
    COMMA = -9
    ASTERISK = -10

    SELECT = 1
    FROM = 2
    WHERE = 3
    AND = 4
    OR = 5
    NOT = 6

    @classmethod
    def get_token(cls, val):
        try:
            # First off, see if we have a valid match to a known keyword
            result = cls[val.upper()]
            if result.value > 0:
                return result
        except:
            pass

        # Lets see if its a symbol
        for ttype, expected in [
            (cls.EQUALS, "="),
            (cls.LT, "<"),
            (cls.LTE, "<="),
            (cls.GT, ">"),
            (cls.GTE, ">="),
            (cls.NE, "<>"),
            (cls.LPAREN, "("),
            (cls.RPAREN, ")"),
            (cls.COMMA, ","),
            (cls.ASTERISK, "*"),
        ]:
            if val == expected:
                return ttype

        # Not a valid keyword, so lets see if it is another valid construct
        for ttype, pat in [
            (cls.REFERENCE, REFERENCE_PAT),
            (cls.NUMBER, NUMBER_PAT),
            (cls.STRING, STRING_PAT),
        ]:
            if pat.fullmatch(val):
                return ttype

        # Unable to recognise token type
        return None

    @classmethod
    def comparators(cls):
        return set([cls.LT, cls.LTE, cls.GT, cls.GTE, cls.EQUALS, cls.NE])

    @classmethod
    def rvalues(cls):
        return set([cls.REFERENCE, cls.STRING, cls.NUMBER])


class _Tokeniser:
    """
    Constructs a tokeniser, which tokenises the given SQL string
    :param sql: SQL string to be tokenised
    :raises InvalidTokenError: Raised if an unrecognised token is read (for example, an invalid keyword)
    """

    def __init__(self, sql: str):
        self._sql = sql
        self._pos = 0  # Use this for tokenising to start with
        self._tokens = [
            t for t in self._tokenise()
        ]  # Note that we tokenise up front rather than using the generator directly to support look ahead
        self._pos = 0  # Then reuse for managing the consumption of tokens

    """
    Consume and return a Token, either simply returning the Token or validating that it is of the expected type. If not, a ValueError is raised
    :param expected: TokenType or set of TokenTypes that we expect to consume
    :returns: Token if present (and matches expected, if given), otherwise None
    :raises ValueError: Raised if consumed token does not match expected type
    """

    def consume(self, expected: typing.Optional[typing.Union[_TokenType, set]] = None):
        if expected and isinstance(expected, _TokenType):
            expected = set([expected])
        if self._pos < len(self._tokens):
            token = self._tokens[self._pos]
            if expected and token.ttype not in expected:
                raise UnexpectedTokenError(token, expected)
            self._pos += 1
            return token
        elif expected:
            raise UnexpectedTokenError(None, expected)
        return None

    """
    Return the next Token without consuming it, so that the parser can determine the appropriate action
    :returns: Next Token that will be consumed
    """

    def peek_next(self):
        return self._tokens[self._pos] if self._pos < len(self._tokens) else None

    """
    Returns whether the next token is of a given type. Like peek_next, does not consume the token
    :returns: True if next token is of the type specified, False otherwise (including if there is no next token)
    """

    def next_is(self, ttype: _TokenType):
        next = self.peek_next()
        return next and next.ttype == ttype

    def _curr_char(self):
        return self._sql[self._pos]

    def _tokenise(self):
        self._pos = 0
        while self._pos < len(self._sql):
            token = ""
            curr = self._curr_char()
            if curr.isspace():
                self._pos += 1
            elif curr == "{":
                # This is a reference, so collect up to and including the closing brace
                token = self._collect("}")
            elif curr in "'\"":
                # This is a string, so collect up to and including the closing matching quote
                token = self._collect(self._sql[self._pos])
            elif curr in SYMBOLS:
                # This is an operator - either a single character or compound comparator, collect the whole operator
                token = self._collect_operator()
            else:
                # Probably a keyword, collect until we hit a separator
                token = self._collect()

            if token:
                ttype = _TokenType.get_token(token)
                if ttype:
                    yield _Token(ttype, token)
                else:
                    raise InvalidTokenError(token)

    def _collect(self, separator=None):
        result = self._sql[self._pos]
        self._pos += 1

        if separator:
            sep_hit = lambda x: x == separator
        else:
            sep_hit = lambda x: x not in ascii_letters + digits + "_."

        while self._pos < len(self._sql) and not sep_hit(self._sql[self._pos]):
            result += self._sql[self._pos]
            self._pos += 1
        if (
            separator
            and self._pos < len(self._sql)
            and self._sql[self._pos] == separator
        ):
            result += separator
            self._pos += 1

        return result

    def _collect_operator(self):
        check = self._sql[self._pos : self._pos + 2]
        if check in [">=", "<=", "<>"]:
            self._pos += 2
            return check
        curr = self._curr_char()
        self._pos += 1
        return curr



from .exceptions import UnexpectedTokenError, UnrecognisedReferenceError
from .tokeniser import _Tokeniser, _TokenType

"""
Supported grammar:
Statement ::= SELECT <References> FROM REFERENCE [WHERE <Where_Clause>]
References ::= ASTERISK | ReferenceList
ReferenceList ::= REFERENCE | REFERENCE COMMA ReferenceList
Where_Clause ::= <Where_Term> [OR Where_Clause]
Where_Term ::= <Where_Factor> [AND <Where_Term>]
Where_Factor ::= <Where_Primary> | NOT <Where_Primary>
Where_Primary ::= LPAREN Where_Clause RPAREN | Condition
Condition ::= REFERENCE COMPARATOR <RValue>
RValue ::= REFERENCE | NUMBER | STRING
"""


def clean_outers(reference):
    # Removes outer curly brackets or quotes from a value
    return reference[1:-1]


class _References:
    def __init__(self):
        self.all_references = False
        self.references = []

    def parse(self, tokeniser):
        if tokeniser.next_is(_TokenType.ASTERISK):
            tokeniser.consume(_TokenType.ASTERISK)
            self.all_references = True
        else:
            self.references.append(tokeniser.consume(_TokenType.REFERENCE).value)
            while tokeniser.next_is(_TokenType.COMMA):
                tokeniser.consume(_TokenType.COMMA)
                self.references.append(tokeniser.consume(_TokenType.REFERENCE).value)

    def filter_fields(self, record):
        if self.all_references:
            return record
        return {key: record[key] for key in map(clean_outers, self.references)}


class _Condition:
    """
    Constructs a condition container as per the grammar above
    """

    def __init__(self):
        self.reference = None
        self.operator = None
        self.rvalue = None

    def parse(self, tokeniser):
        self.reference = tokeniser.consume(_TokenType.REFERENCE).value
        self.operator = tokeniser.consume(_TokenType.comparators())
        self.rvalue = tokeniser.consume(_TokenType.rvalues())

    def __repr__(self):
        return " ".join([self.reference, self.operator.value, self.rvalue.value])

    def satisfied(self, record):
        _Condition._validate_reference(self.reference, record)
        if self.rvalue.ttype == _TokenType.REFERENCE:
            _Condition._validate_reference(self.rvalue.value, record)
        lvalue = record[clean_outers(self.reference)]
        match self.rvalue.ttype:
            case _TokenType.REFERENCE:
                rvalue = record[clean_outers(self.rvalue.value)]
            case _TokenType.STRING:
                rvalue = clean_outers(self.rvalue.value)
            case _:
                rvalue = self.rvalue.value
        if not isinstance(lvalue, str):
            rvalue = type(lvalue)(rvalue)

        match self.operator.ttype:
            case _TokenType.LT:
                return lvalue < rvalue
            case _TokenType.LTE:
                return lvalue <= rvalue
            case _TokenType.GT:
                return lvalue > rvalue
            case _TokenType.GTE:
                return lvalue >= rvalue
            case _TokenType.EQUALS:
                return lvalue == rvalue
            case _TokenType.NE:
                return lvalue != rvalue

    @staticmethod
    def _validate_reference(reference, record):
        if not clean_outers(reference) in record:
            raise UnrecognisedReferenceError(reference)


class _WherePrimary:
    """
    Constructs a where primary container as per the grammar above
    """

    def __init__(self):
        self.where_clause = None
        self.condition = None

    def parse(self, tokeniser):
        if tokeniser.next_is(_TokenType.LPAREN):
            tokeniser.consume(_TokenType.LPAREN)
            self.where_clause = _WhereClause()
            self.where_clause.parse(tokeniser)
            tokeniser.consume(_TokenType.RPAREN)
        else:
            self.condition = _Condition()
            self.condition.parse(tokeniser)

    def __repr__(self):
        if self.where_clause:
            return "( " + repr(self.where_clause) + " )"
        else:
            return repr(self.condition)

    def satisfied(self, record):
        if self.where_clause:
            return self.where_clause.satisfied(record)
        else:
            return self.condition.satisfied(record)


class _WhereFactor:
    """
    Constructs a where factor container as per the grammar above
    """

    def __init__(self):
        self.where_primary = _WherePrimary()
        self.bool_not = False

    def parse(self, tokeniser):
        if tokeniser.next_is(_TokenType.NOT):
            tokeniser.consume(_TokenType.NOT)
            self.bool_not = True
        self.where_primary.parse(tokeniser)

    def __repr__(self):
        return ("NOT " if self.bool_not else "") + repr(self.where_primary)

    def satisfied(self, record):
        return self.where_primary.satisfied(record) ^ self.bool_not


class _WhereTerm:
    """
    Constructs a where term container as per the grammar above
    """

    def __init__(self):
        self.where_factor = _WhereFactor()
        self.where_term = None

    def parse(self, tokeniser):
        self.where_factor.parse(tokeniser)
        if tokeniser.next_is(_TokenType.AND):
            tokeniser.consume(_TokenType.AND)
            self.where_term = _WhereTerm()
            self.where_term.parse(tokeniser)

    def __repr__(self):
        return repr(self.where_factor) + (
            " AND " + repr(self.where_term) if self.where_term else ""
        )

    def satisfied(self, record):
        return self.where_factor.satisfied(record) and (
            self.where_term is None or self.where_term.satisfied(record)
        )


class _WhereClause:
    """
    Constructs a where clause container as per the grammar above
    """

    def __init__(self):
        self.where_term = _WhereTerm()
        self.where_clause = None

    def parse(self, tokeniser):
        self.where_term.parse(tokeniser)
        if tokeniser.next_is(_TokenType.OR):
            tokeniser.consume(_TokenType.OR)
            self.where_clause = _WhereClause()
            self.where_clause.parse(tokeniser)

    def __repr__(self):
        return repr(self.where_term) + (
            " OR " + repr(self.where_clause) if self.where_clause else ""
        )

    def satisfied(self, record):
        return self.where_term.satisfied(record) or (
            self.where_clause is not None and self.where_clause.satisfied(record)
        )


class _Parser:
    """
    Constructs a parser and parses the given SQL, storing the reference and conditions to be used when querying data
    :param sql: SQL to be parsed
    :raises InvalidTokenException: Raised (by the tokeniser) if an invalid token is read
    :raises UnexpectedTokenException: Raised when parsing we hit a token which does not match expected type
    """

    def __init__(self, sql: str):
        self.tokeniser = _Tokeniser(sql)
        # Because references and fromref are straightforward, we store them directly here, but as the
        # where clause hierarchy is more complex, that is stored in child objects
        self._references = _References()
        self._fromref = ""
        self._where_clause = None
        self._parse()

    def satisfied(self, record):
        return (
            True if self._where_clause is None else self._where_clause.satisfied(record)
        )

    def filter_fields(self, record):
        return self._references.filter_fields(record)

    def from_ref(self):
        return clean_outers(self._fromref)

    def _parse(self):
        self.tokeniser.consume(_TokenType.SELECT)
        self._parse_references()
        self.tokeniser.consume(_TokenType.FROM)
        self._fromref = self.tokeniser.consume(_TokenType.REFERENCE).value
        if self.tokeniser.next_is(_TokenType.WHERE):
            self.tokeniser.consume(_TokenType.WHERE)
            self._parse_where_clause()
        if not self.tokeniser.peek_next() is None:
            raise UnexpectedTokenError(self.tokeniser.peek_next())

    def _parse_references(self):
        self._references.parse(self.tokeniser)

    def _parse_where_clause(self):
        self._where_clause = _WhereClause()
        self._where_clause.parse(self.tokeniser)
